/*
 * KOF91 GP32 port
 *
 * CopyLeft 2003-2004 by THOR / Shadow Of Dawn
 *
 * All this code is unprotected: do what you want with it !
 *
 * Becareful : as it is a lazy port, implementations are often limited to the use of the game.
 *
 * Code to read .flc files. Really limited to the use I've found in the game, really
 * not a complete FLC reader. If you need more, ask for it !
 *
 */

#include "fli.h"
#include "system.h"
#include "gfx.h"

#include <stdlib.h>

DATAFILE*	fli_df;
BITMAP*		fli_bitmap;
BITMAP*		fli_second;
COLOR		fli_palette[256];

#define CEL_DATA    3
#define COLOR_256   4
#define DELTA_FLC   7
#define COLOR_64   11
#define DELTA_FLI  12
#define BLACK      13
#define BYTE_RUN   15
#define FLI_COPY   16
#define PSTAMP     18
#define DTA_BRUN   25
#define DTA_COPY   26
#define DTA_LC     27
#define LABEL      31
#define BMP_MASK   32
#define MLEV_MASK  33
#define SEGMENT    34
#define KEY_IMAGE  35
#define KEY_PAL    36
#define REGION     37
#define WAVE       38
#define USERSTRING 39
#define RGN_MASK   40
#define LABELEX    41
#define SHIFT      42
#define PATHMAP    43
                                         
#define PREFIX_TYPE   0xF100
#define SCRIPT_CHUNK  0xF1E0
#define FRAME_TYPE    0xF1FA
#define SEGMENT_TABLE 0xF1FB
#define HUFFMAN_TABLE 0xF1FC

/* FLC header */
                                         
int		size;          /* Size of FLIC including this header */
short	type;          /* File type 0xAF11, 0xAF12, 0xAF30, 0xAF44, ... */
short	frames;        /* Number of frames in first segment */
short	width;         /* FLIC width in pixels */
short	height;        /* FLIC height in pixels */
short	depth;         /* Bits per pixel (usually 8) */
short	flags;         /* Set to zero or to three */
int		speed;         /* Delay between frames */
short	reserved1;     /* Set to zero */
int		created;       /* Date of FLIC creation (FLC only) */
int		creator;       /* Serial number or compiler id (FLC only) */
int		updated;       /* Date of FLIC update (FLC only) */
int		updater;       /* Serial number (FLC only), see creator */
short	aspect_dx;     /* Width of square rectangle (FLC only) */
short	aspect_dy;     /* Height of square rectangle (FLC only) */
short	ext_flags;     /* EGI: flags for specific EGI extensions */
short	keyframes;     /* EGI: key-image frequency */
short	totalframes;   /* EGI: total number of frames (segments) */
int		req_memory;    /* EGI: maximum chunk size (uncompressed) */
short	max_regions;   /* EGI: max. number of regions in a CHK_REGION chunk */
short	transp_num;    /* EGI: number of transparent levels */
ubyte	reserved2[24]; /* Set to zero */
int		oframe1;       /* Offset to frame 1 (FLC only) */
int		oframe2;       /* Offset to frame 2 (FLC only) */
ubyte	reserved3[40]; /* Set to zero */

/* Frame header */

int		f_size;			/* Size of the chunk, including subchunks */
int		f_type;			/* Chunk type: 0xF1FA */
short	f_chunks;		/* Number of subchunks */
short	f_delay;		/* Delay in milliseconds */
short	f_reserved;		/* Always zero */
short	f_width;		/* Frame width override (if non-zero) */
short	f_height;		/* Frame height override (if non-zero) */

/* Working data */

int				offset;
signed char*	buf;
int				nbcol;

int fli_btoi(signed char b)
{
	int a = b;
	return (a<0?256+a:a);
}

int fli_readInt()
{
	int res = fli_btoi(buf[offset]) + (fli_btoi(buf[offset+1])<<8) + (fli_btoi(buf[offset+2])<<16) + (fli_btoi(buf[offset+3])<<24);
	offset+=4;
	return res;
}

short fli_readShort()
{
	int res = fli_btoi(buf[offset]) + (fli_btoi(buf[offset+1])<<8);
	offset+=2;
	return (short)res;
}

void fli_readBytes(ubyte* t,int l)
{
	int i;
	for (i=0;i<l;i++)
		t[i] = buf[offset++];
}

void fli_readPalette(short type)
{
	int n = fli_readShort(),i,j=0,k,copy,r,g,b;
	for (i=0;i<n;i++)
	{
		j += fli_btoi(buf[offset++]); /* skip */
		copy = fli_btoi(buf[offset++]);
		if (copy==0)
			copy = 256;
		for (k=0;k<copy;k++)
		{
			r = fli_btoi(buf[offset++]);
			g = fli_btoi(buf[offset++]);
			b = fli_btoi(buf[offset++]);
			if (type==COLOR_64)
			{
				r<<=2;
				g<<=2;
				b<<=2;
			}
			fli_palette[j++] = makecol(r,g,b);
		}
	}
}

void fli_readRLE()
{
	int i,j,k,l;
	signed char b;
	for (i=0;i<height;i++)
	{
		offset++; /* ignore packet count */
		for(j=0;j<width;)
		{
			b = buf[offset++];
			if (b<0)
			{
				for (l=-b;l>0;l--)
					fli_bitmap->buf[i+(j++)*height] = fli_palette[fli_btoi(buf[offset++])];
			}
			else
			{
				k = fli_palette[fli_btoi(buf[offset++])];
				for (l=0;l<b;l++)
					fli_bitmap->buf[i+(j++)*height] = k;
			}
		}
	}
}

void fli_readDeltaFLC()
{
	short nbl = fli_readShort(); /* nb of lines */
	int nbs = 0,i,opcode,last,j,c,l,k0,k1,tmp;
	signed char b;
	for (i=0;i<nbl;i++)
	{
		last=-1;
		while (((opcode = fli_readShort())&0xc000)!=0)
		{
			if ((opcode&0xc000)==0xc000)
				nbs-=opcode;
			else if ((opcode&0x8000)==0x8000)
				last = opcode&0xff;
			else
				GpAppExit();
		}
		c = 0;
		for (j=0;j<opcode;j++)
		{
			c += fli_btoi(buf[offset++]); /* col skip */
			b = buf[offset++]; /* RLE */
			if (b>0)
			{
				for (l=b;l>0;l--)
				{
					tmp = (i+nbs)+(c++)*height;
					fli_second->buf[tmp] = fli_bitmap->buf[tmp] = fli_palette[fli_btoi(buf[offset++])];
					tmp = (i+nbs)+(c++)*height;
					fli_second->buf[tmp] = fli_bitmap->buf[tmp] = fli_palette[fli_btoi(buf[offset++])];
				}
			}
			else if (b<0)
			{
				k0 = fli_palette[fli_btoi(buf[offset++])];
				k1 = fli_palette[fli_btoi(buf[offset++])];
				for (l=0;l<-b;l++)
				{
					tmp = (i+nbs)+(c++)*height;
					fli_second->buf[tmp] = fli_bitmap->buf[tmp] = k0;
					tmp = (i+nbs)+(c++)*height;
					fli_second->buf[tmp] = fli_bitmap->buf[tmp] = k1;
				}
			}
		}
/*
		if (last>=0)
			fli_bitmap->buf[(i+nbs+1)*width-1] = fli_palette[last];
*/
	}
}

void fli_readDeltaFLI()
{
	short nbs = fli_readShort(); /* skipped lines */
	short nbl = fli_readShort(); /* nb of lines */
	int i,j,nbp,c,k,l,tmp;
	signed char b;
	for (i=0;i<nbl;i++)
	{
		nbp = fli_btoi(buf[offset++]); /* nb of packets in the line */
		c = 0;
		for (j=0;j<nbp;j++)
		{
			c += fli_btoi(buf[offset++]); /* col skip */
			b = buf[offset++]; /* RLE */
			if (b>0)
			{
				for (l=b;l>0;l--)
				{
					tmp = (i+nbs)+(c++)*height;
					fli_second->buf[tmp] = fli_bitmap->buf[tmp] = fli_palette[fli_btoi(buf[offset++])];
				}
			}
			else if (b<0)
			{
				k = fli_palette[fli_btoi(buf[offset++])];
				for (l=0;l<-b;l++)
				{
					tmp = (i+nbs)+(c++)*height;
					fli_second->buf[tmp] = fli_bitmap->buf[tmp] = k;
				}
			}
		}
	}
}

void fli_readChunk()
{
	int c_size = fli_readInt();
	int c_end  = offset+c_size-4; 
	int c_type = fli_readShort()&0xffff;
	switch(c_type)
	{
		case CEL_DATA:
			break;
		case COLOR_256:
			fli_readPalette(COLOR_256);
			break;
		case DELTA_FLC:
			fli_readDeltaFLC();
			break;
		case COLOR_64:
			fli_readPalette(COLOR_64);
			break;
		case DELTA_FLI:
			fli_readDeltaFLI();
			break;
		case BLACK:
			break;
		case BYTE_RUN:
			fli_readRLE();
			break;
		case FLI_COPY:
			break;
		case PSTAMP:
			break;
		case DTA_BRUN:
			break;
		case DTA_COPY:
			break;
		case DTA_LC:
			break;
		case LABEL:
			break;
		case BMP_MASK:
			break;
		case MLEV_MASK:
			break;
		case SEGMENT:
			break;
		case KEY_IMAGE:
			break;
		case KEY_PAL:
			break;
		case REGION:
			break;
		case WAVE:
			break;
		case USERSTRING:
			break;
		case RGN_MASK:
			break;
		case LABELEX:
			break;
		case SHIFT:
			break;
		case PATHMAP:
			break;

		case PREFIX_TYPE:
			break;
		case SCRIPT_CHUNK:
			break;
		case FRAME_TYPE:
			break;
		case SEGMENT_TABLE:
			break;
		case HUFFMAN_TABLE:
			break;
		default:
			GpAppExit();
	}
	offset = c_end;
}

int open_fli(const char* name)
{
	fli_df = open_datafile(name,0);
	read_datafile(fli_df,132);
	buf = (signed char*)fli_df->dat;
	offset = 0;
	size = fli_readInt();
	type = fli_readShort();
	frames = fli_readShort();
	width = fli_readShort();
	height = fli_readShort();
	depth = fli_readShort();
	flags = fli_readShort();
	speed = fli_readInt();
	reserved1 = fli_readShort();
	created = fli_readInt();
	creator = fli_readInt();
	updated = fli_readInt();
	updater = fli_readInt();
	aspect_dx = fli_readShort();
	aspect_dy = fli_readShort();
	ext_flags = fli_readShort();
	keyframes = fli_readShort();
	totalframes = fli_readShort();
	req_memory = fli_readInt();
	max_regions = fli_readShort();
	transp_num = fli_readShort();
	fli_readBytes(reserved2,sizeof(reserved2));
	oframe1 = fli_readInt();
	oframe2 = fli_readInt();
	fli_readBytes(reserved3,sizeof(reserved3));
	fli_second = fli_bitmap = create_bitmap(width,height);
	return FLI_OK;
}

void reset_fli_variables()
{
}

int next_fli_frame(int skip)
{
	int i;

	if (offset==size)
		offset = oframe2;

	f_size		= fli_readInt();
	if (fli_df->read==offset)
		read_datafile(fli_df,(ulong)f_size);
	f_type		= fli_readShort()&0xffff;
	f_chunks	= fli_readShort();
	f_delay		= fli_readShort();
	f_reserved	= fli_readShort();
	f_width		= fli_readShort();
	f_height	= fli_readShort();

	if (f_type!=FRAME_TYPE)
		GpAppExit();

	for (i=0;i<f_chunks;i++)
		fli_readChunk();

	return offset<size?FLI_OK:0;
}

void close_fli()
{
	if (fli_df==NULL)
		return;
	unload_datafile(fli_df);
	destroy_bitmap(fli_bitmap);
	fli_bitmap = NULL;
	fli_df = NULL;
}

void fli_screen(BITMAP *dst)
{
	fli_second = dst;
}
