/*
 * KOF91 GP32 port
 *
 * CopyLeft 2003-2004 by THOR / Shadow Of Dawn
 *
 * All this code is unprotected: do what you want with it !
 *
 * Becareful : as it is a lazy port, implementations are often limited to the use of the game.
 *
 * Sprites and things like that, the "Becareful" remark takes really its place here :)
 *
 */

#include "gfx.h"
#include "system.h"

#include "gpmem.h"
#include "gpgraphic.h"
#include "gpstdlib.h"

#include <string.h>

#define ASM_GFX

BITMAP* screen = NULL;
PALLETE black_palette;
FONT* font = NULL;

int trans;

ubyte nflip;
GPDRAWSURFACE gpDraw[2]; /* buffers */

/* drawing mode vars */
int dm_mode = DRAW_MODE_SOLID;
BITMAP* dm_bitmap;
int dm_x,dm_y;

extern BITMAP* zoomscreen;
extern BITMAP* virtscreen;

#define CACHE_SIZE 128

int nb_cache = 0, cache_enabled = 0;
BITMAP* bm_cache[CACHE_SIZE];
RLE_SPRITE* rle_cache[CACHE_SIZE];

ubyte gamma[256];

void gp32_initGfx()
{
	int i,j,k;
	GpGraphicModeSet(GPC_DFLAG_16BPP, NULL);
	nflip = 1;
    GpLcdSurfaceGet(&gpDraw[0], 0);
    GpLcdSurfaceGet(&gpDraw[1], 1);
	gm_memset(gpDraw[0].ptbuffer,0,gpDraw[0].buf_w*gpDraw[0].buf_h*2);
	gm_memset(gpDraw[1].ptbuffer,0,gpDraw[1].buf_w*gpDraw[1].buf_h*2);
	GpSurfaceSet(&gpDraw[0]);
	/*virtscreen->buf = (COLOR*)gpDraw[nflip].ptbuffer;*/
	k=0;
	for (i=0;i<256;i++)
	{
		j = (i+k<248?i+k:248);
		if (i%2)
			k+=((i<128)?1:-1);
		gamma[i] = (j==248)?j:((j&0xf8)+(((j&0x07)>3)?8:0));
	}
	trans=makecol(255,0,255);
}

void gp32_flip()
{
	GpSurfaceFlip(&gpDraw[nflip]);
	nflip ^= 1;
	/*virtscreen->buf = (COLOR*)gpDraw[nflip].ptbuffer;*/
}

extern void blit_KOF91();

void gp32_error(char*msg)
{
	if (font!=NULL)
	{
		textout(virtscreen,font,msg,0,0);
		blit_KOF91();
		while(buttonpressed());
		while(!buttonpressed());
	}
	GpAppExit();
}

BITMAP* create_bitmap(int w,int h)
{
	BITMAP* b = gm_malloc(sizeof(BITMAP));
	b->w = w;
	b->h = h;
	b->buf = gm_malloc(w*h*sizeof(COLOR));
	return b;
}

COLOR makecol(ushort r,ushort g,ushort b)
{
	return (gamma[r]<<8)|(gamma[g]<<3)|(gamma[b]>>2);
}

short readShort(ubyte* buf,ulong* offset)
{
	short s = buf[*offset+1];
	s = (s<<8)+buf[(*offset)];
	*offset+=2;
	return s;
}

BITMAP* load_pcx (const char* name,PALLETE pal)
{
	DATAFILE *df;
	ulong offset = 4,palette;
    int x,y,w,h,bpl,bpl3;
	int i,j,k,l;
	ubyte b,nbp;
	BITMAP *pcx;
	ubyte *line;

	if ((strcmp(name+strlen(name)-4,".pcx")&&strcmp(name+strlen(name)-4,".PCX")) || (df = load_datafile(name))==NULL)
		return NULL;

	x = readShort(df->dat,&offset);
    y = readShort(df->dat,&offset);
	w = readShort(df->dat,&offset)-x+1;
    h = readShort(df->dat,&offset)-y+1;

	offset+=53;
	nbp = df->dat[offset++];
	bpl = readShort(df->dat,&offset);
	offset+=60;
	bpl3 = bpl*nbp;
    line = gm_malloc(bpl3);
	palette = df->len-768;
	pcx = create_bitmap(w,h);
	for (i=0;i<h;i++)
	{
    	for (j=0;j<bpl3;)
    	{
			b = df->dat[offset++];
			if ((b&0xc0)==0xc0)
			{
				l = b&0x3f;
				b = df->dat[offset++];
				for (k=0;k<l;k++)
					line[j++] = b;
			}
			else
				line[j++] = b;
	    }
		for (j=0;j<w;j++)
		{
			if (nbp==1)
			{
				l = palette+line[j]*3;
				pcx->buf[j*h+i] = makecol(df->dat[l],df->dat[l+1],df->dat[l+2]);
			}
			else
				pcx->buf[j*h+i] = makecol(line[j],line[j+bpl],line[j+bpl+bpl]);
		}
	}
	gm_free(line);
	unload_datafile(df);
	return pcx;
}

int nb_save = 0;
RECT save_area[12];

void init_bkgd_save()
{
	nb_save = 0;
}

#define MIN(a,b) ((a)<(b)?(a):(b))

extern void asm_draw_sprite(ushort*,ushort*,int,int);

void _draw_sprite(BITMAP* dst,BITMAP* src,int x,int y,int sx,int sy,int svbg,int flip)
{
	int ssx = x<0?sx-x:sx;
	int ssy = y<0?sy-y:sy;
	int i,h=src->h-ssy,w,j;
	COLOR *s = src->buf + (flip ? 0 : (src->h*ssx)+ssy);
	COLOR *d;

	if (x<0)
		x = 0;
	if (y<0)
		y = 0;
 	d = dst->buf+(x*dst->h)+y;
	if (h+y>dst->h)
		h = dst->h-y;
	w = src->w-ssx;
	if (x+w>dst->w)
		w = dst->w-x;
	if (flip)
		s += src->h*(((w<src->w && x==0) ? w : src->w)-1);

#ifndef ASM_GFX
	for (i=0;i<w;++i)
	{
		for (j=0;j<h;++j)
		{
			if (s[j]!=trans)
				d[j] = s[j];
		}
		if (flip)
			s-=src->h;
		else
			s+=src->h;
		d+=dst->h;
	}
#else
	if (w<=0 || h<=0)
		return;
	asm_draw_sprite(s,d,(((flip?-src->h:src->h)<<1)<<16)|(dst->h<<1),(w<<16)|(h<<1));
#endif
	if (svbg && dst==zoomscreen)
	{
		for (i=0;i<nb_save;i++)
		{
			if (   (x>=save_area[i].x) && (x+w<save_area[i].x+save_area[i].w)
				&& (y>=save_area[i].y) && (y+h<save_area[i].y+save_area[i].h))
				return;
		}
		save_area[nb_save].x = x;
		save_area[nb_save].y = y;
		save_area[nb_save].w = w;
		save_area[nb_save].h = h;
		nb_save+=1;
	}
}

void draw_sprite(BITMAP* dst,BITMAP* src,int x,int y)
{
	_draw_sprite(dst,src,x,y,0,0,1,0);
}

void draw_sprite_h_flip(BITMAP* dst,BITMAP* src,int x,int y) 
{
	_draw_sprite(dst,src,x,y,0,0,1,1);
}

void _blit(BITMAP* src,BITMAP* dst,int sx,int sy,int dx,int dy,int dw,int dh)
{
	int i,h=src->h-sy;
	COLOR *s = src->buf+(sx*src->h)+sy;
	COLOR *d = dst->buf+(dx*dst->h)+(dy>0?dy:0);
	if (dy<0)
	{
		h+=dy;
		s-=dy;
	}
	else if (h+dy>dst->h)
		h = dst->h-dy;
	if (h>dh)
		h = dh;
	h *= sizeof(COLOR);
	for (i=sx;i<src->w && dx<dst->w && dw>0;i++)
	{
		if (dx++>=0)
			gm_memcpy(d,s,h);
		dw-=1;
		s+=src->h;
		d+=dst->h;
	}
}

void blit_screen(COLOR* src,ulong* dst)
{
	int i,j,k;
	src+=320*240;
	dst+=(319*120)-1;
	for (i=0;i<320;++i)
	{
		for (j=0;j<24;j++)
		{
			k = *(--src);
			*(++dst) = k|((*(--src))<<16);
			k = *(--src);
			*(++dst) = k|((*(--src))<<16);
			k = *(--src);
			*(++dst) = k|((*(--src))<<16);
			k = *(--src);
			*(++dst) = k|((*(--src))<<16);
			k = *(--src);
			*(++dst) = k|((*(--src))<<16);
		}
		dst-=240;
	}
}

extern void asm_blit_screen(ushort*,ulong*);

void blit(BITMAP* src,BITMAP* dst,int sx,int sy,int dx,int dy,int dw,int dh)
{
	if (src==dst && dx>0)
	{
		BITMAP *b = clone_bitmap(src);
		_blit(b,dst,sx,sy,dx,dy,dw,dh);
		destroy_bitmap(b);
	}
	else if (dst==NULL)
	{
#ifndef ASM_GFX
		blit_screen(src->buf,(ulong*)gpDraw[nflip].ptbuffer);
#else
		asm_blit_screen(src->buf,(ulong*)gpDraw[nflip].ptbuffer);
#endif
	}
	else
		_blit(src,dst,sx,sy,dx,dy,dw,dh);
}

void masked_blit(BITMAP* src,BITMAP* dst,int sx,int sy,int dx,int dy,int sw,int sh)
{
	_draw_sprite(dst,src,dx,dy,sx,sy,1,0);
}

void rectfill(BITMAP* b,int x0,int y0,int x1,int y1,COLOR c)
{
	int i,j,w=x1-x0,h=y1-y0;
	if (dm_mode==DRAW_MODE_SOLID)
	{
		COLOR* dst = b->buf+(x0*b->h)+y0;
		for (i=0;i<w;i++)
		{
			for (j=0;j<h;j++)
				dst[j] = c;
			dst+=b->h;
		}
	}
	else
	{
		for (i=(dm_x>0?dm_x-dm_bitmap->w:0);i<b->w;i+=dm_bitmap->w)
		{
			for (j=(dm_y>0?dm_y-dm_bitmap->h:0);j<b->h;j+=dm_bitmap->h)
				blit(dm_bitmap,b,0,0,i,j,x1,y1);
		}
	}
}

void textout(BITMAP* b,FONT* f,const char* text,int x,int y)
{
	int i,j, l = strlen(text);
	BITMAP *tmp = create_bitmap(f->width,f->bmp->h);
	for (i=0;i<l;i++)
	{
		for (j=0;j<f->charset->len && f->charset->dat[j]!=text[i];j++);
		if (j==f->charset->len)
			j=0;
		_blit(f->bmp,tmp,j*tmp->w,0,0,0,tmp->w,tmp->h);
		_draw_sprite(b,tmp,x+i*(f->width+1),y,0,0,1,0);
	}
	destroy_bitmap(tmp);
}

void textout_centre(BITMAP* b,FONT* f,const char* text,int x,int y,COLOR c)
{
	textout(b,f,text,x-(strlen(text)*(f->width+1)/2),y-(f->bmp->h/2));
}

void clear(BITMAP* b)
{
	if (b==NULL)
		gm_memset(gpDraw[nflip].ptbuffer,0,320*240*sizeof(COLOR));
	else
		gm_memset(b->buf,0,b->w*b->h*sizeof(COLOR));
}

COLOR getpixel(BITMAP* b,int x,int y)
{
	return b->buf[x*b->h+y];
}

void putpixel(BITMAP* b,int x,int y,COLOR c)
{
	b->buf[x*b->h+y] = c;
}

void drawing_mode(int mode,BITMAP* b,int x,int y)
{
	dm_mode = mode;
	dm_bitmap = b;
	dm_x = x;
	dm_y = y;
}

void rect(BITMAP* b,int x0,int y0,int x1,int y1,COLOR c)
{
	int i,j,k,w=x1-x0,h=y1-y0;
	/* horizontal */
	j = (x0+0)*b->h+y0;
	for (i=0;i<w;i++)
	{
		b->buf[j] = c;
		b->buf[j+h] = c;
		j+=b->h;
	}

	/* vertical */
	j = (x0+0)*b->h+y0+1;
	k = (x1+0)*b->h+y0+1;
	for (i=1;i<h;i++)
		b->buf[j++] = b->buf[k++] = c;
}

int yy[1024];

extern void asm_stretch_blit(ushort*,ushort*,int,int);

void stretch_blit(BITMAP* src,BITMAP* dst,int sx,int sy,int sw,int sh,int dx,int dy,int dw,int dh)
{
	COLOR *s = src->buf+(sx*src->h)+sy;
	COLOR *d = dst->buf+(dx*dst->h)+dy;
	int i,j,k;
	int ddx = (sw<<8)/dw;
	int ddy = ((sh>src->h?src->h:sh)<<8)/dh;

	for (i=0;i<dh;i++)
		yy[i] = ((i*ddy)>>8);

#ifdef ASM_GFX
	if (src!=zoomscreen || dst!=virtscreen)
	{
#endif
		for (i=0;i<dw;i++)
		{
			k = ((i*ddx)>>8)*src->h;
			for (j=0;j<dh;j++)
				d[j] = s[k+yy[j]];
			d+=dst->h;
		}
#ifdef ASM_GFX
	}
	else
	{
		for (i=0;i<dh;i++)
			yy[i] <<= 1;
		asm_stretch_blit(s,d,ddx,(src->h<<1));
	}
#endif
}

void stretch_sprite(BITMAP* dst,BITMAP* src,int x,int y,int w,int h) 
{
	COLOR *s = src->buf;
	COLOR *d = dst->buf+(x*dst->h)+y;
	int i,j,k,l;
	int dx,dy;
	if (w==0||h==0)
		return;
	dx = (src->w<<8)/w;
	dy = (src->h<<8)/h;
	for (i=0;i<h;i++)
		yy[i] = ((i*dy)>>8);
	for (i=0;i<w;i++)
	{
		k = ((i*dx)>>8)*src->h;
		for (j=0;j<h;j++)
		{
			if (s[k+yy[j]]!=trans)
				d[j] = s[k+yy[j]];
		}
		d+=dst->h;
	}
}

void destroy_bitmap(BITMAP* b)
{
	gm_free(b->buf);
	gm_free(b);
}

BITMAP* load_bitmap(const char* name,int i)
{
	return load_pcx(name,NULL);
}

BITMAP*	clone_bitmap(BITMAP* b)
{
	BITMAP *cpy = create_bitmap(b->w,b->h);
	gm_memcpy(cpy->buf,b->buf,b->w*b->h*sizeof(COLOR));
	return cpy;
}

void enable_rle_cache()
{
	nb_cache = 0;
	cache_enabled = 1;
}

void disable_rle_cache()
{
	int i;
	if (!cache_enabled)
		return;
	for (i=0;i<nb_cache;i++)
	{
		rle_cache[i]->cache_idx = -1;
		rle_cache[i]->cache_hits = 0;
		destroy_bitmap(bm_cache[i]);
	}
	nb_cache = 0;
	cache_enabled = 0;
}

#define CACHE_LIM 1000000

void put_cache(RLE_SPRITE* rle,BITMAP *b)
{
	int i,h,hi;
	int asize = gm_availablesize();

	if (!cache_enabled)
	{
		destroy_bitmap(b);
		return;
	}

	rle->cache_hits+=1;
	if ((nb_cache<CACHE_SIZE) && (asize-rle->size>CACHE_LIM))
	{
		rle->cache_idx		= nb_cache;
		rle_cache[nb_cache] = rle;
		bm_cache[nb_cache++]= b;
		return;
	}
	h = rle->cache_hits*rle->size;
	for (i=0;i<nb_cache;i++)
	{
		hi = rle_cache[i]->cache_hits*rle_cache[i]->size;
		if ((hi<h) && (asize+rle_cache[i]->size-rle->size)>CACHE_LIM)
		{
			destroy_bitmap(bm_cache[i]);
			rle_cache[i]->cache_idx	= -1;
			bm_cache[i]				= b;
			rle_cache[i]			= rle;
			rle->cache_idx			= i;
			return;
		}
	}
	/* no place in cache */
	destroy_bitmap(b);
}

void asm_unpack(ubyte*,ushort*,ushort*,ushort*);

void draw_rle_sprite(BITMAP* b,RLE_SPRITE* s,int x,int y,int flip)
{
	BITMAP *rle;
	ubyte *src;
	COLOR *dst, *end;
#ifndef ASM_GFX
	int i=0,j;
	ubyte n;
	COLOR c;
#endif

	if (s->cache_idx>=0)
	{
		_draw_sprite(b,bm_cache[s->cache_idx],x+(flip?s->z:s->x),y+s->y,0,0,1,flip);
		s->cache_hits+=1;
		return;
	}

	rle = create_bitmap(s->w,s->h);
	dst = rle->buf;
	src = s->buf;
	end = dst+(s->w*s->h);

#ifndef ASM_GFX
	while (dst<end)
	{
		n = src[i++];
		if ((n&0x80)==0x80)
		{
			n&=0x7f;
			c = s->pal[src[i++]];
			for (j=0;j<n;++j)
				*(dst++) = c;
		}
		else
		{
			for (j=0;j<n;++j)
				*(dst++) = s->pal[src[i++]];
		}
	}
#else
	asm_unpack(src,dst,end,s->pal);
#endif
	_draw_sprite(b,rle,x+(flip?s->z:s->x),y+s->y,0,0,s->svbg,flip);
	put_cache(s,rle);
}

RLE_SPRITE*	get_rle_sprite(BITMAP* b)
{
	RLE_SPRITE *spr = (RLE_SPRITE*)gm_malloc(sizeof(RLE_SPRITE));
	ubyte *temp,*dst,*s0,*s1,*end,*buf = (ubyte*)gm_malloc(b->w*b->h);
	COLOR c,*pal = (COLOR*)gm_malloc(256*sizeof(COLOR));
	ubyte n,cc;
	int minx=b->w,miny=b->h,maxx=0,maxy=0,i,j,k;

	spr->nbc = 0;
	pal[spr->nbc++] = trans;
	for (i=0;i<b->w;i++)
	{
		for (j=0;j<b->h;j++)
		{
			c = b->buf[i*b->h+j];
			k=0;
			if (c!=trans)
			{
				if (i<minx)
					minx = i;
				if (j<miny)
					miny = j;
				if (i>maxx)
					maxx = i;
				if (j>maxy)
					maxy = j;
				while (k<spr->nbc && pal[k]!=c)
					++k;
				if (k==spr->nbc)
				{
					if (k<256)
						pal[spr->nbc++] = c;
					else
						k = 0;
				}
			}
			buf[i*b->h+j]=k;
		}
	}
	spr->pal = (COLOR*)gm_malloc(spr->nbc*sizeof(COLOR));
	gm_memcpy(spr->pal,pal,spr->nbc*sizeof(COLOR));
	gm_free(pal);

	spr->x = minx;
	spr->y = miny;
	spr->z = b->w-maxx;
	if ((minx>0)||(miny>0)||(maxx+1<b->w)||(maxy+1<b->h))
	{
		spr->w = maxx-minx+1;
		spr->h = maxy-miny+1;
		temp = (ubyte*)gm_malloc(spr->w*spr->h);
		for (i=0;i<spr->w;i++)
			for (j=0;j<spr->h;j++)
				temp[i*spr->h+j] = buf[(i+minx)*b->h+j+miny];
		gm_free(buf);
		buf = temp;
	}
	else
	{
		spr->w = b->w;
		spr->h = b->h;
	}

	dst = temp = (ubyte*)gm_malloc(spr->w*spr->h*2);
	s0 = s1 = buf;
	end = buf+(spr->w*spr->h);
	n = 0;
	while (s1!=end)
	{
		cc = *(s1++);
		++n;
		if (s1==end)
			break;
		if (cc!=*s1)
		{
			if (n<127)
				continue;
			*(dst++) = n;
			gm_memcpy(dst,s0,n);
			dst += n;
			s0 = s1;
			n = 0;
		}
		else
		{
			if (n>1)
			{
				*(dst++) = --n;
				gm_memcpy(dst,s0,n);
				dst += n;
				n = 1;
			}
			while ((s1!=end) && (cc==*s1) && (n<127))
			{
				++n;
				++s1;
			}
			*(dst++) = n|0x80;
			*(dst++) = cc;
			s0 = s1;
			n = 0;
		}
	}
	if (n>0)
	{
		*(dst++) = n;
		gm_memcpy(dst,s0,n);
		dst += n;
	}
	gm_free(buf);
	spr->buf = (ubyte*)gm_malloc(dst-temp);
	gm_memcpy(spr->buf,temp,dst-temp);
	gm_free(temp);
	spr->size = spr->w*spr->h*sizeof(COLOR);
	spr->cache_idx  = -1;
	spr->cache_hits = 0;
	spr->svbg = 1;
	spr->ow = b->w;
	spr->oh = b->h;
	return spr;
}

void destroy_rle_sprite(RLE_SPRITE* s)
{
	gm_free(s->pal);
	gm_free(s->buf);
	gm_free(s);
}

RLE_SPRITE* frgd_spr[64];
int frgd_nb = 0;

void set_foreground(BITMAP *frgd)
{
	int w,h;
	COLOR *buf;
	int i,j,k,l;
	int a,b,c,d;
	BITMAP* tmp;

	frgd_nb = 0;
	if (frgd==NULL)
		return;
	i=0;
	w = frgd->w;
	h = frgd->h;
	buf = frgd->buf;
	while (i<w)
	{
		/* find first non empty column */
		for (j=i;j<w;j++)
		{
			for (l=0;l<h && buf[j*h+l]==trans;l++);
			if (l<h)
				break;
		}
		if (j==w)
			break;
		/* find first empty column */
		for (k=j+1;(k<w)&&((k-j)<64);k++)
		{
			for (l=0;l<h && buf[k*h+l]==trans;l++);
			if (l==h)
				break;
		}
		/* horizontal cut */
		a = 0;
		while (a<h)
		{
			/* find first non empty line */
			d=k;
			for (b=a;b<h && d==k;b++)
				for (d=j;d<k && buf[d*h+b]==trans;d++)
					if (0) gp32_error("bad 1");
			if (b==h)
			{
				b = 0;
				c = h;
			}
			else
			{
				/* find first empty line */
				d=0;
				for (c=b+1;c<h && d!=k;c++)
					for (d=j;d<k && buf[d*h+c]==trans;d++)
						if (0) gp32_error("bad 2");
			}
			/* get foreground part */
			tmp = create_bitmap(k-j,c-b);
			_blit(frgd,tmp,j,b,0,0,tmp->w,tmp->h);
			frgd_spr[frgd_nb]       = get_rle_sprite(tmp);
			frgd_spr[frgd_nb]->x    = j;
			frgd_spr[frgd_nb]->y  = b;
			frgd_spr[frgd_nb++]->svbg = 0;
			destroy_bitmap(tmp);
			a = c+1;
		}
		i = k;
	}
}

void show_foreground(int x,int y,int w,int h)
{
	int i;
	for (i=0;i<frgd_nb;i++)
	{
		if (	(frgd_spr[i]->x>=x+w)
			 || ((frgd_spr[i]->x+frgd_spr[i]->w)<x)
			 || (frgd_spr[i]->y>=y+h)
			 || ((frgd_spr[i]->y+frgd_spr[i]->h)<y))
			continue;
		draw_rle_sprite(zoomscreen,frgd_spr[i],0,0,0);
	}
}

void clear_foreground()
{
	int i;
	for (i=0;i<frgd_nb;i++)
		destroy_rle_sprite(frgd_spr[i]);
	frgd_nb = 0;
}

void restore_background(BITMAP* b)
{
	int i;
	if (!nb_save)
	{
		blit ( b, zoomscreen , 0 , 0, 0, 0, b->w, b->h ) ;
		return;
	}
	for (i=0;i<nb_save;i++)
		_blit(b,zoomscreen,save_area[i].x,save_area[i].y,save_area[i].x,save_area[i].y,save_area[i].w,save_area[i].h);
	nb_save = 0;
}

void acquire_screen()
{
/**
	char buf[64];
	gm_sprintf(buf,"%ld",gm_availablesize());
	if (font!=NULL)
		textout(virtscreen, font, buf, 0, 0);
**/
}

void release_screen()
{
	gp32_flip();
}

FONT* load_font(const char*name)
{
	FONT *f = (FONT*)gm_malloc(sizeof(FONT));
	char buf[64];
	strcpy(buf,name);
	strcat(buf,".txt");
	f->charset = load_datafile(buf);
	strcpy(buf,name);
	strcat(buf,".pcx");
	f->bmp = load_pcx(buf,NULL);
	f->width = f->bmp->w/f->charset->len;
	return f;
}

void destroy_font(FONT* f)
{
	destroy_bitmap(f->bmp);
	unload_datafile(f->charset);
	gm_free(f);
	if (font==f)
		font = NULL;
}
