#include "lib.h"

int errno;

#define MAX_FILES 32
FILE files[MAX_FILES];
int free_file = -1;

FILE* fopen(const char* path,const char* mode)
{
	FILE* f;
	if (free_file<0)
	{
		gm_memset(files,0,sizeof(files));
		free_file = 0;
	}
	while (files[free_file].handle!=0)
	{
		free_file+=1;
		if (free_file==MAX_FILES)
			free_file = 0;
	}
	f = &files[free_file];
	f->offset = f->end = 0;
	if (mode[0]=='r')
		errno = GpFileOpen(path,OPEN_R,&f->handle);
	else
		errno = GpFileCreate(path,ALWAYS_CREATE,&f->handle);
	return errno == SM_OK ? f : NULL;
}

#define MIN(a,b) ((a)<(b) ? (a) : (b))

ulong fread(void* buf,int isize,int icount,FILE* f)
{
	ulong rd;
	int l = isize*icount, bsize;
	ubyte* _buf = (ubyte*)buf;
	if (f->offset==f->end)
	{
		f->offset = 0;
		GpFileRead(f->handle,f->data,l>BUFSIZE?BUFSIZE2:BUFSIZE,&rd);
		f->end = (int)rd;
	}
	bsize = f->end-f->offset;
	if (l<=bsize)
	{
		gm_memcpy(_buf,f->data+f->offset,l);
		f->offset+=l;
		errno = 0;
		return icount;
	}
	gm_memcpy(_buf,f->data+f->offset,bsize);
	f->offset = f->end = 0;
	errno = GpFileRead(f->handle,_buf+bsize,l-bsize,&rd);
	if (errno==ERR_EOF)
		errno = 0;
	return (rd+bsize)/isize;
}

int ferror()
{
	return errno;
}

int fseek(FILE* f,long pos,int from)
{
	long current;
	switch(from)
	{
		case SEEK_END: from = FROM_END; break;
		case SEEK_SET: from = FROM_BEGIN; break;
		case SEEK_CUR: from = FROM_CURRENT; break;
	}
	if (from == FROM_CURRENT)
	{
		if (f->offset+pos<f->end)
		{
			f->offset += pos;
			return 0;
		}
		pos -= f->end-f->offset;
	}
	errno = GpFileSeek(f->handle,from,pos,&current);
	f->end = f->offset = 0;
	return errno!=SM_OK ? 1 : 0;
}

long ftell(FILE* f)
{
	long current;
	errno = GpFileSeek(f->handle,0,FROM_CURRENT,&current);
	return current-(f->end-f->offset);
}

int fclose(FILE* f)
{
	GpFileClose(f->handle);
	f->handle = 0;
	return 0;
}

FILE* fdopen(int fd,const char* mode)
{
	return NULL;
}
