#include "options.h"
#include "gfx.h"
#include "system.h"
#include "sound.h"
#include "config.h"
#include "gpstdlib.h"

#include <string.h>

extern BITMAP *virtscreen;
extern BITMAP *screen;

void Bing (void);
void Done (void);
void blit_KOF91 (void);

#define NB_OPTIONS 14

/* default game options */
const int def_options[NB_OPTIONS] =
{
	0,	/* Music ON */
	0,	/* SFX ON */
	0,	/* BACKGROUNDS ANIM ON */
	1,	/* FOREGROUNS OFF */
	0,	/* FIGHT ANIMATIONS ON */
	1,	/* RANDOM BKGD IN STORY MODE OFF */
	0,	/* SHOW HELP IN MAIN MENU */
	3,	/* CPU SPEED 133 */
	1,	/* FRAMESKIP 1 */
	0,	/* WPUNCH A */
	1,	/* SPUNCH B */
	3,	/* WKICK R */
	2,	/* SKICK L */
	0	/* SKIN */
};

const char* menu[NB_OPTIONS] =
{
	"MUSIC",
	"SOUND FX",
	"BKGRD ANIMS",
	"FOREGROUNDS",
	"INTRO/ENDING",
	"RND STORY",
	"SHOW CONFIG",
	"OVERCLOCK",
	"FRAMESKIP",
	"WPUNCH",
	"SPUNCH",
	"WKICK",
	"SKICK",
	"SKIN"
};

char* items[] =
{
	"ON","OFF",
	"ON","OFF",
	"ON","OFF",
	"ON","OFF",
	"ON","OFF",
	"ON","OFF",
	"ON","OFF",
	"80","100","120","133","144","150","156","166",
	"0","1","2","3",
	"A","B","L","R",
	"A","B","L","R",
	"A","B","L","R",
	"A","B","L","R",
	"CLASSIC",NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL
};

int icount[NB_OPTIONS] = {2,2,2,2,2,2,2,8,4,4,4,4,4,1};

const char eol[] = {0xd,0xa,0};

int options[NB_OPTIONS];

extern BITMAP *virtscreen;
extern FONT* font;

void drawMenu(const char**menu,char**choices,int*count,int*sel,int x,int y,int w,int n,int idx,char*msg,int xm,int ym)
{
	int i,j=0;
	clear(virtscreen);
	for (i=0;i<n;i++)
	{
		if (idx==i)
			textout(virtscreen,font,">",x-10,y);
		textout(virtscreen,font,menu[i],x,y);
		textout(virtscreen,font,choices[j+sel[i]],x+w,y);
		j+=count[i];
		y+=12;
	}
	textout(virtscreen,font,msg,xm,ym);
	blit_KOF91();
}

void loadOptions()
{
	char opt[16],optnum[3];
	int num;
	override_config_file("config.ini");
	for (num=0;num<NB_OPTIONS;num++)
	{
		gcvt ( num , num<10 ? 1 : 2 , optnum ) ;
		strcpy(opt,"option");
		strcat(opt,optnum);
		options[num] = get_config_int(NULL,opt,def_options[num]);
	}
}

void saveOptions()
{
	char optnum[3],buf[1024];
	int num;
	buf[0] = 0;
	for (num=0;num<NB_OPTIONS;num++)
	{
		strcat(buf,"option");
		gcvt ( num , num<10 ? 1 : 2 , optnum ) ;
		strcat(buf,optnum);
		strcat(buf," = ");
		gcvt ( options[num] , 1 , optnum ) ;
		strcat(buf,optnum);
		strcat(buf,eol);
	}
	save_file("config.ini",(ubyte*)buf,strlen(buf));
}

void configMenu()
{
	int key,idx = 0,n,mod = 1,i,tmp;

	Done();
	while (1)
	{
		if (mod)
		{
			drawMenu(menu,items,icount,options,70,30,130,NB_OPTIONS,idx,"START: BACK TO GAME",70,210);
			GpThreadSleep(300);
			mod = 0;
		}
		key = GpKeyGet();
		if (key & GPC_VK_START)
			break;
		if (idx<OPTION_WPUNCH || idx>OPTION_SKICK)
		{
			if (key & GPC_VK_FB)
			{
				options[idx] = options[idx]==0?icount[idx]-1:options[idx]-1;
				mod = 1;
			}
			else if ((key & GPC_VK_FA)||(key & GPC_VK_SELECT))
			{
				options[idx] = options[idx]+1<icount[idx]?options[idx]+1:0;
				mod = 1;
			}
		}
		else
		{
			if (key & (GPC_VK_FA|GPC_VK_FB|GPC_VK_FL|GPC_VK_FR))
			{
				tmp = options[idx];
				options[idx] = (key & GPC_VK_FA) ? 0 : ((key & GPC_VK_FB) ? 1 : ((key & GPC_VK_FL) ? 2 : 3));
				for (i=OPTION_WPUNCH;i<=OPTION_SKICK;i++)
				{
					if (i!=idx && options[i]==options[idx])
						options[i] = tmp;
				}
				mod = 1;
			}
		}
		n = idx;
		if (((key&GPC_VK_UP)||(key&GPC_VK_LEFT)))
			n = (n>0 ? n : NB_OPTIONS)-1;
		if (((key&GPC_VK_DOWN)||(key&GPC_VK_RIGHT)) && n<NB_OPTIONS-1)
			n = (n+1)%NB_OPTIONS;
		if (n!=idx || mod)
		{
			Bing();
			idx = n;
			mod = 1;
		}
	}
	Done();
	while(buttonpressed());
	saveOptions();
}

int getOptionValue(int o)
{
	return options[o];
}

void setOptionValue(int o,int v)
{
	options[o] = v;
}

void addChoice(int o,char* v)
{
	int i,j=0;
	for (i=0;i<o;i++)
		j+=icount[i];
	items[j+(icount[o]++)] = v;
}
