/*
 * KOF91 GP32 port
 *
 * CopyLeft 2003-2004 by THOR / Shadow Of Dawn
 *
 * All this code is unprotected: do what you want with it !
 *
 * Becareful : as it is a lazy port, implementations are often limited to the use of the game.
 *
 * This part of code use the MODPLAY library by CHN.
 *
 */

#include "sound.h"

#include "system.h"

#include "gpdef.h"
#include "gpmm.h"
#include "gpstdlib.h"

#include <string.h>

#define SND_RATE 22050
#define MONO

void gp32_initSound()
{
	GpPcmStop();
#ifdef MONO
	GpPcmInit(PCM_M22,PCM_16BIT);
#else
	GpPcmInit(PCM_S22,PCM_16BIT);
#endif
}

char music_enabled = 1;
char sfx_enabled = 1;

void setMusicEnabled(char b)
{
	music_enabled = b;
}

void setSFXEnabled(char b)
{
	sfx_enabled = b;
}

MUSIC* load_mod(char* name)
{
	int i;
	MUSIC *music = (MUSIC*)gm_malloc(sizeof(MUSIC));
	if (!music_enabled)
	{
		music->mod = NULL;
		return music;
	}
	music->mod = (MODPlay*)gm_malloc(sizeof(MODPlay));
	MODPlay_Init(music->mod);
	music->df = load_datafile(name);
	for (i=0;i<31;i++)
	{
		/* CHN lib bug ? seems like right chanel does not accept volume adjustment...always max. */
		/* no volume set fonction : need to decrease volume before playing for each sample */
		ubyte v = music->df->dat[45+i*30]>>1;
		music->df->dat[45+i*30] = v==0?1:v;
	}
	MODPlay_SetMOD(music->mod,music->df->dat);
	MODPlay_SetPlayFreq(music->mod,SND_RATE);
#ifdef MONO
	MODPlay_SetStereo(music->mod,0);
#endif
	return music;
}

void play_mod(MUSIC* music)
{
	if (music_enabled && music->mod)
		MODPlay_Start(music->mod);
}

void destroy_mod(MUSIC* music)
{
	if (music->mod)
	{
		MODPlay_Stop(music->mod);
		unload_datafile(music->df);
		gm_free(music->mod);
	}
	gm_free(music);
}

void play_sample(SAMPLE* sample)
{
	if (sfx_enabled && sample->buf)
	{
		sample->play = 1;
		GpPcmPlay((ushort*)sample->buf,sample->len,0);
	}
}

int readInt(ubyte *buf)
{
	return buf[0]|(buf[1]<<8)|(buf[2]<<16)|(buf[3]<<24);
}

SAMPLE* _load_sample(const char* name)
{
	DATAFILE *df = load_datafile(name);
	int rate = readInt(df->dat+24);
	ubyte bytes = df->dat[32];
	ubyte bits = df->dat[34];
	int length;
	SAMPLE *s = (SAMPLE*)gm_malloc(sizeof(SAMPLE));
	ubyte *src;
	ushort tmp;
	int i,j=(SND_RATE<<8)/rate,k=(rate<<8)/SND_RATE,l,ii,is;
	/* each sample need to have the same freq/format */
	i=36;
	while (df->dat[i]!='d'||df->dat[i+1]!='a'||df->dat[i+2]!='t'||df->dat[i+3]!='a')
		++i;
	length = readInt(df->dat+i+4);
	src = df->dat+i+8;
#ifdef MONO
	s->len = ((length/bytes)*j)>>7;
	l = (s->len>>1);
#else
	s->len = ((length/bytes)*j)>>6;
	l = (s->len>>2);
#endif
	s->buf = (ubyte*)gm_malloc(s->len);
	ii = 0;
	for (i=0;i<l;i++)
	{
		is = ((k*i)>>8)*bytes;
		if (bits==16)
		{
			if (bytes==4) /* already 16b stereo */
			{
				tmp = (src[is]|(src[is+1]<<8))+0x8000;
				s->buf[ii++] = (tmp&0xff);
				s->buf[ii++] = (tmp>>8);
#ifndef MONO
				tmp = (src[is]|(src[is+1]<<8))+0x8000;
				s->buf[ii++] = (tmp&0xff);
				s->buf[ii++] = (tmp>>8);
#endif
			}
			else
			{
				tmp = (src[is]|(src[is+1]<<8))+0x8000;
				s->buf[ii++] = (tmp&0xff);
				s->buf[ii++] = (tmp>>8);
#ifndef MONO
				s->buf[ii++] = (tmp&0xff);
				s->buf[ii++] = (tmp>>8);
#endif
			}
		}
		else
		{
			s->buf[ii++] = 00;
			s->buf[ii++] = src[is];
#ifndef MONO
			s->buf[ii++] = 00;
			s->buf[ii++] = src[is+(bytes==2?1:0)];
#endif
		}
	}
	unload_datafile(df);
	s->play = 0;
	return s;
}

SAMPLE* load_sample(const char* name)
{
	SAMPLE *s;
	if (sfx_enabled)
		return _load_sample(name);
	s = (SAMPLE*)gm_malloc(sizeof(SAMPLE));
	s->buf = NULL;
	s->len = 0;
	s->play = 0;
	return s;
}

void destroy_sample(SAMPLE* sample)
{
	if (sample->buf)
	{
		if (sample->play)
			stop_sample(sample);
		gm_free(sample->buf);
	}
	gm_free(sample);
}

void stop_sample(SAMPLE* sample)
{
	if (sample->play && sample->buf)
		GpPcmRemove((unsigned short*)sample->buf);
	sample->play = 0;
}
