/*
 * KOF91 GP32 port
 *
 * CopyLeft 2003-2004 by THOR / Shadow Of Dawn
 *
 * All this code is unprotected: do what you want with it !
 *
 * Becareful : as it is a lazy port, implementations are often limited to the use of the game.
 *
 */

#include "system.h"
#include "options.h"

#include "gpdef.h"
#include "gpmem.h"
#include "gpstdio.h"
#include "gpstdlib.h"

#include "unzip.h"

#include <string.h>

JOYSTICK joy[2];
ubyte key[127];
int key_pressed;

char zip_name[32];
unzFile zip_uf = NULL;
unz_global_info zip_gi;

void EnableCache();

void gp32_initSystem()
{
	GpFatInit();
	GpRelativePathSet("gp:\\GPMM\\KOF91");
	EnableCache();
	gp32_set_speed(SPEED_133);
}

const int oc_freq[] = {     78,    105,    123,    135,    144,    150,    156,    165};
const int oc_div[]  = {0x2c002,0x1b001,0x21001,0x25001,0x28001,0x2a001,0x2c001,0x2f001};

/* yes, 133Mhz. I'm too lazy to learn ASM atm */
void gp32_set_speed(int speed)
{
	GpClockSpeedChange(oc_freq[speed]*1000000,oc_div[speed],2);
}

void clear_keybuf()
{
	gm_memset(key,0,sizeof(key));
	key_pressed = 0;
}

const int BUTTONS[] = {GPC_VK_FA,GPC_VK_FB,GPC_VK_FL,GPC_VK_FR};

void poll_joystick()
{
	int k = GpKeyGet();
	gm_memset(joy,0,sizeof(joy));
	if (k & BUTTONS[getOptionValue(OPTION_SPUNCH)])
		joy[0].button[0].b = 1;
	if (k & BUTTONS[getOptionValue(OPTION_WPUNCH)])
		joy[0].button[1].b = 1;
	if (k & BUTTONS[getOptionValue(OPTION_SKICK)])
		joy[0].button[2].b = 1;
	if (k & BUTTONS[getOptionValue(OPTION_WKICK)])
		joy[0].button[3].b = 1;
	if (k & GPC_VK_UP)
		joy[0].stick[0].axis[1].d1 = 1;
	if (k & GPC_VK_DOWN)
		joy[0].stick[0].axis[1].d2 = 1;
	if (k & GPC_VK_LEFT)
		joy[0].stick[0].axis[0].d1 = 1;
	if (k & GPC_VK_RIGHT)
		joy[0].stick[0].axis[0].d2 = 1;

	if ((k & GPC_VK_START) && (k & GPC_VK_FL))
		GpAppExit();
	else if (k & GPC_VK_SELECT)
		key[KEY_ESC]=1;
	else if (k & GPC_VK_START)
		key[KEY_SPACE]=1;
}

int buttonpressed()
{
	int k = GpKeyGet();
	return
       ((k & GPC_VK_FA) ||
	    (k & GPC_VK_FB) ||
	    (k & GPC_VK_FL) ||
	    (k & GPC_VK_FR) ||
	    (k & GPC_VK_START) ||
	    (k & GPC_VK_SELECT));
}

int keypressed()
{
	int k = key_pressed;
	key_pressed = 0;
	if (k!=0)
		return k;
	k = GpKeyGet();
	if (k & GPC_VK_START)
		k = KEY_SPACE;
	else if (k & GPC_VK_SELECT)
		k = KEY_ESC;
	else if (k & GPC_VK_FR)
        k = KEY_F5;
	else
		k = 0;
	return k;
}

void simulate_keypress(int k)
{
	key_pressed = k;
}

/* FILE SYSTEM */

DATAFILE* open_datafile(const char* name,ubyte load)
{
	DATAFILE *df = gm_malloc(sizeof(DATAFILE));

	if (zip_uf!=NULL)
	{
		df->handle = 0;
		df->len = 0;
		df->read = 0;
		df->dat = readZip(name,&df->len);
		if (df->dat==NULL)
		{
			gm_free(df);
			df = NULL;
		}
		return df;
	}

	if (GpFileGetSize(name,&df->len)!=SM_OK)
		return NULL;
	df->dat = gm_malloc(df->len);
	GpFileOpen(name,OPEN_R,&df->handle);
	if (load)
	{
		GpFileRead(df->handle,df->dat,df->len,&df->read);
		GpFileClose(df->handle);
		df->handle = 0;
	}
	else
		df->read = 0;
	return df;
}

DATAFILE* load_datafile(const char* name)
{
	return open_datafile(name,1);
}

void read_datafile(DATAFILE *df,ulong count)
{
	ulong read;
	GpFileRead(df->handle,df->dat+df->read,count,&read);
	df->read += read;
}

void unload_datafile(DATAFILE* file)
{
	if (file==NULL)
		return;
	if (file->handle)
		GpFileClose(file->handle);
	gm_free(file->dat);
	gm_free(file);
}

void save_file(char* name,ubyte* buf,ulong len)
{
	F_HANDLE handle;
	GpFileCreate(name,ALWAYS_CREATE,&handle);
	GpFileWrite(handle,buf,len);
	GpFileClose(handle);
}

/*
 * File research
 */

int _al_file_isok(const char* s)
{
	return 1;
}

GPDIRENTRY *_al_list;
ulong _al_nb;
char _al_folder[64],_al_ext[5];
ulong _al_lext,_al_idx;
GPFILEATTR _al_attr;
int _al_type;

/* pattern is allways : <dir_name>\*<[.]extension> */
void* _al_findfirst(const char* pattern,int type,char* name,int* unused)
{
	int i=0;
	ulong nbr;

	*unused = 0;
	_al_type = type;

	/* find folder name */
	for(i=0;pattern[i]!='\\';i++)
		_al_folder[i] = pattern[i];
	_al_folder[i] = 0;

	/* get file extension */
	strcpy(_al_ext,pattern+i+2);
	strupr(_al_ext);
	_al_lext = strlen(_al_ext);

	GpDirEnumNum(_al_folder,&_al_nb);
	_al_list = (GPDIRENTRY*)gm_malloc(sizeof(GPDIRENTRY)*_al_nb);
	GpDirEnumList(_al_folder,0,_al_nb,_al_list,&_al_nb);
	_al_idx = 0;
	_al_findnext(_al_list,name,unused);
	if (_al_idx==_al_nb)
	{
		_al_findclose(_al_list);
		return NULL;
	}
	else
		return _al_list;
}

void replace_filename(char* buf,char* pattern,char* name,int bufsize)
{
	strcpy(buf,name);
}

int _al_findnext(void* entries,char* name,int* unused)
{
	int i;
	for (;_al_idx<_al_nb;_al_idx++)
	{
		strcpy(name,_al_folder);
		strcat(name,"\\");
		strcat(name,_al_list[_al_idx].name);
		if (_al_type==FA_DIREC)
		{
			GpFileAttr(name,&_al_attr);
			if (_al_attr.attr&0x10) /* directory ? */
				break;
		}
		else
		{
			i = strlen(_al_list[_al_idx].name)-_al_lext;
			if (strcmp(_al_list[_al_idx].name+i,_al_ext)==0)
				break;
		}
	}
	if (_al_idx==_al_nb)
	{
		name[0] = 0;
		return 1;
	}
	else
	{
		_al_idx+=1;
		return 0;
	}
}

void _al_findclose(void* entries)
{
	gm_free(_al_list);
}

void gcvt(int num,int size,char* buf)
{
	gm_sprintf(buf,"%d",num);
}

int openZip(const char* name)
{
	if (*name==0)
		return 0;
	closeZip();
	zip_uf = unzOpen(name);
	if (zip_uf==NULL)
		return 1;
	strcpy(zip_name,name);
	if (unzGetGlobalInfo (zip_uf,&zip_gi)!=UNZ_OK)
		return 2;
	return 0;
}

void closeZip()
{
	if (zip_uf==NULL)
		return;
	unzClose(zip_uf);
	zip_uf = NULL;
}

ubyte *readZip(const char *name,ulong *len)
{
	unz_file_info info;
	ubyte *buf;
	char fname[32];
	int i,j=0;

	/* test if initial format : sprites\<char>\<file> */
	for (i=0;name[i];i++)
		if (name[i]=='\\')
			j++;

	if (j==2)
		strcpy(fname,name+8);
	else
		strcpy(fname,name);

	/* new format : <char>/<file> */
	for (i=0;fname[i] && fname[i]!='\\';i++);
	if (fname[i])
		fname[i] = '/';

	if (unzLocateFile(zip_uf,fname,2)!=UNZ_OK)
	{
		/* if <char>/<file> not found, search <file> */
		if (unzLocateFile(zip_uf,fname+i+1,2)!=UNZ_OK)
			return NULL;
	}
	unzGetCurrentFileInfo(zip_uf,&info,NULL,0,NULL,0,NULL,0);
	if (unzOpenCurrentFile(zip_uf)!=UNZ_OK)
		return NULL;
	buf = (ubyte*)gm_malloc(info.uncompressed_size);
	if (unzReadCurrentFile(zip_uf,buf,info.uncompressed_size)<=0)
	{
		gm_free(buf);
		buf = NULL;
	}
	*len = info.uncompressed_size;
	unzCloseCurrentFile(zip_uf);
	return buf;
}
